@@ -53,31 +53,30 @@ impl Array {
     }
 
     /// Concatenate two `Array`s.
-    pub fn concatenate(left: &Array, right: &Array) -> Result<Array> {
+    pub fn concatenate(left: &Array, right: &Array) -> Array {
         use Array::*;
         match (left, right) {
-            (Bool(l), Bool(r)) => Ok(Bool(ArrayExt::concatenate(l, r))),
+            (Bool(l), Bool(r)) => Bool(ArrayExt::concatenate(l, r)),
 
-            (F32(l), F32(r)) => Ok(F32(ArrayExt::concatenate(l, r))),
-            (F64(l), F64(r)) => Ok(F64(ArrayExt::concatenate(l, r))),
+            (F32(l), F32(r)) => F32(ArrayExt::concatenate(l, r)),
+            (F64(l), F64(r)) => F64(ArrayExt::concatenate(l, r)),
 
-            (C32(l), C32(r)) => Ok(C32(ArrayExt::concatenate(l, r))),
-            (C64(l), C64(r)) => Ok(C64(ArrayExt::concatenate(l, r))),
+            (C32(l), C32(r)) => C32(ArrayExt::concatenate(l, r)),
+            (C64(l), C64(r)) => C64(ArrayExt::concatenate(l, r)),
 
-            (I16(l), I16(r)) => Ok(I16(ArrayExt::concatenate(l, r))),
-            (I32(l), I32(r)) => Ok(I32(ArrayExt::concatenate(l, r))),
-            (I64(l), I64(r)) => Ok(I64(ArrayExt::concatenate(l, r))),
+            (I16(l), I16(r)) => I16(ArrayExt::concatenate(l, r)),
+            (I32(l), I32(r)) => I32(ArrayExt::concatenate(l, r)),
+            (I64(l), I64(r)) => I64(ArrayExt::concatenate(l, r)),
 
-            (U8(l), U8(r)) => Ok(U8(ArrayExt::concatenate(l, r))),
-            (U16(l), U16(r)) => Ok(U16(ArrayExt::concatenate(l, r))),
-            (U32(l), U32(r)) => Ok(U32(ArrayExt::concatenate(l, r))),
-            (U64(l), U64(r)) => Ok(U64(ArrayExt::concatenate(l, r))),
+            (U8(l), U8(r)) => U8(ArrayExt::concatenate(l, r)),
+            (U16(l), U16(r)) => U16(ArrayExt::concatenate(l, r)),
+            (U32(l), U32(r)) => U32(ArrayExt::concatenate(l, r)),
+            (U64(l), U64(r)) => U64(ArrayExt::concatenate(l, r)),
 
-            (l, r) => Err(error(format!(
-                "Cannot concatenate arrays with different data types: {}, {}",
-                l.dtype(),
-                r.dtype()
-            ))),
+            (l, r) if l.dtype() > r.dtype() => Array::concatenate(l, &r.cast_into(l.dtype())),
+            (l, r) if l.dtype() < r.dtype() => Array::concatenate(&l.cast_into(r.dtype()), r),
+
+            _ => unreachable!(),
         }
     }
 
@@ -122,7 +121,7 @@ impl Array {
             C32(_) => ComplexType::C32.into(),
             C64(_) => ComplexType::C32.into(),
             F32(_) => FloatType::F32.into(),
-            F64(_) => FloatType::F32.into(),
+            F64(_) => FloatType::F64.into(),
             I16(_) => IntType::I16.into(),
             I32(_) => IntType::I32.into(),
             I64(_) => IntType::I64.into(),
@@ -252,58 +251,92 @@ impl Array {
 
     /// Element-wise equality comparison.
     pub fn eq(&self, other: &Array) -> Array {
+        use Array::*;
+        match (self, other) {
+            (Bool(l), Bool(r)) => Bool(l.eq(r.deref())),
+            (C32(l), C32(r)) => Bool(l.eq(r.deref())),
+            (C64(l), C64(r)) => Bool(l.eq(r.deref())),
+            (F32(l), F32(r)) => Bool(l.eq(r.deref())),
+            (F64(l), F64(r)) => Bool(l.eq(r.deref())),
+            (I16(l), I16(r)) => Bool(l.eq(r.deref())),
+            (I32(l), I32(r)) => Bool(l.eq(r.deref())),
+            (I64(l), I64(r)) => Bool(l.eq(r.deref())),
+            (U8(l), U8(r)) => Bool(l.eq(r.deref())),
+            (U16(l), U16(r)) => Bool(l.eq(r.deref())),
+            (U32(l), U32(r)) => Bool(l.eq(r.deref())),
+            (U64(l), U64(r)) => Bool(l.eq(r.deref())),
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l.eq(&r.cast_into(l_dtype)),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => l.cast_into(r_dtype).eq(r),
+                (l, r) => unreachable!("{} equal to {}", l, r),
+            },
+        }
+    }
+
+    /// Raise `e` to the power of `self`.
+    pub fn exp(&self) -> Array {
         use Array::*;
         match self {
-            Bool(l) => Bool(l.eq(&other.type_cast())),
-            C32(l) => Bool(l.eq(&other.type_cast())),
-            C64(l) => Bool(l.eq(&other.type_cast())),
-            F32(l) => Bool(l.eq(&other.type_cast())),
-            F64(l) => Bool(l.eq(&other.type_cast())),
-            I16(l) => Bool(l.eq(&other.type_cast())),
-            I32(l) => Bool(l.eq(&other.type_cast())),
-            I64(l) => Bool(l.eq(&other.type_cast())),
-            U8(l) => Bool(l.eq(&other.type_cast())),
-            U16(l) => Bool(l.eq(&other.type_cast())),
-            U32(l) => Bool(l.eq(&other.type_cast())),
-            U64(l) => Bool(l.eq(&other.type_cast())),
+            Bool(l) => l.exp().into(),
+            C32(l) => l.exp().into(),
+            C64(l) => l.exp().into(),
+            F32(l) => l.exp().into(),
+            F64(l) => l.exp().into(),
+            I16(l) => l.exp().into(),
+            I32(l) => l.exp().into(),
+            I64(l) => l.exp().into(),
+            U8(l) => l.exp().into(),
+            U16(l) => l.exp().into(),
+            U32(l) => l.exp().into(),
+            U64(l) => l.exp().into(),
         }
     }
 
     /// Element-wise greater-than comparison.
     pub fn gt(&self, other: &Array) -> Array {
         use Array::*;
-        match self {
-            Bool(l) => Bool(l.gt(&other.type_cast())),
-            C32(l) => Bool(l.gt(&other.type_cast())),
-            C64(l) => Bool(l.gt(&other.type_cast())),
-            F32(l) => Bool(l.gt(&other.type_cast())),
-            F64(l) => Bool(l.gt(&other.type_cast())),
-            I16(l) => Bool(l.gt(&other.type_cast())),
-            I32(l) => Bool(l.gt(&other.type_cast())),
-            I64(l) => Bool(l.gt(&other.type_cast())),
-            U8(l) => Bool(l.gt(&other.type_cast())),
-            U16(l) => Bool(l.gt(&other.type_cast())),
-            U32(l) => Bool(l.gt(&other.type_cast())),
-            U64(l) => Bool(l.gt(&other.type_cast())),
+        match (self, other) {
+            (Bool(l), Bool(r)) => Bool(l.gt(r.deref())),
+            (C32(l), C32(r)) => Bool(l.gt(r.deref())),
+            (C64(l), C64(r)) => Bool(l.gt(r.deref())),
+            (F32(l), F32(r)) => Bool(l.gt(r.deref())),
+            (F64(l), F64(r)) => Bool(l.gt(r.deref())),
+            (I16(l), I16(r)) => Bool(l.gt(r.deref())),
+            (I32(l), I32(r)) => Bool(l.gt(r.deref())),
+            (I64(l), I64(r)) => Bool(l.gt(r.deref())),
+            (U8(l), U8(r)) => Bool(l.gt(r.deref())),
+            (U16(l), U16(r)) => Bool(l.gt(r.deref())),
+            (U32(l), U32(r)) => Bool(l.gt(r.deref())),
+            (U64(l), U64(r)) => Bool(l.gt(r.deref())),
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l.gt(&r.cast_into(l_dtype)),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => l.cast_into(r_dtype).gt(r),
+                (l, r) => unreachable!("{} greater than {}", l, r),
+            },
         }
     }
 
     /// Element-wise greater-or-equal comparison.
     pub fn gte(&self, other: &Array) -> Array {
         use Array::*;
-        match self {
-            Bool(l) => Bool(l.gte(&other.type_cast())),
-            C32(l) => Bool(l.gte(&other.type_cast())),
-            C64(l) => Bool(l.gte(&other.type_cast())),
-            F32(l) => Bool(l.gte(&other.type_cast())),
-            F64(l) => Bool(l.gte(&other.type_cast())),
-            I16(l) => Bool(l.gte(&other.type_cast())),
-            I32(l) => Bool(l.gte(&other.type_cast())),
-            I64(l) => Bool(l.gte(&other.type_cast())),
-            U8(l) => Bool(l.gte(&other.type_cast())),
-            U16(l) => Bool(l.gte(&other.type_cast())),
-            U32(l) => Bool(l.gte(&other.type_cast())),
-            U64(l) => Bool(l.gte(&other.type_cast())),
+        match (self, other) {
+            (Bool(l), Bool(r)) => Bool(l.gte(r.deref())),
+            (C32(l), C32(r)) => Bool(l.gte(r.deref())),
+            (C64(l), C64(r)) => Bool(l.gte(r.deref())),
+            (F32(l), F32(r)) => Bool(l.gte(r.deref())),
+            (F64(l), F64(r)) => Bool(l.gte(r.deref())),
+            (I16(l), I16(r)) => Bool(l.gte(r.deref())),
+            (I32(l), I32(r)) => Bool(l.gte(r.deref())),
+            (I64(l), I64(r)) => Bool(l.gte(r.deref())),
+            (U8(l), U8(r)) => Bool(l.gte(r.deref())),
+            (U16(l), U16(r)) => Bool(l.gte(r.deref())),
+            (U32(l), U32(r)) => Bool(l.gte(r.deref())),
+            (U64(l), U64(r)) => Bool(l.gte(r.deref())),
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l.gte(&r.cast_into(l_dtype)),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => l.cast_into(r_dtype).gte(r),
+                (l, r) => unreachable!("{} greater than or equal to {}", l, r),
+            },
         }
     }
 
@@ -348,58 +381,72 @@ impl Array {
     /// Element-wise less-than comparison.
     pub fn lt(&self, other: &Array) -> Array {
         use Array::*;
-        match self {
-            Bool(l) => Bool(l.lt(&other.type_cast())),
-            C32(l) => Bool(l.lt(&other.type_cast())),
-            C64(l) => Bool(l.lt(&other.type_cast())),
-            F32(l) => Bool(l.lt(&other.type_cast())),
-            F64(l) => Bool(l.lt(&other.type_cast())),
-            I16(l) => Bool(l.lt(&other.type_cast())),
-            I32(l) => Bool(l.lt(&other.type_cast())),
-            I64(l) => Bool(l.lt(&other.type_cast())),
-            U8(l) => Bool(l.lt(&other.type_cast())),
-            U16(l) => Bool(l.lt(&other.type_cast())),
-            U32(l) => Bool(l.lt(&other.type_cast())),
-            U64(l) => Bool(l.lt(&other.type_cast())),
+        match (self, other) {
+            (Bool(l), Bool(r)) => Bool(l.lt(r.deref())),
+            (C32(l), C32(r)) => Bool(l.lt(r.deref())),
+            (C64(l), C64(r)) => Bool(l.lt(r.deref())),
+            (F32(l), F32(r)) => Bool(l.lt(r.deref())),
+            (F64(l), F64(r)) => Bool(l.lt(r.deref())),
+            (I16(l), I16(r)) => Bool(l.lt(r.deref())),
+            (I32(l), I32(r)) => Bool(l.lt(r.deref())),
+            (I64(l), I64(r)) => Bool(l.lt(r.deref())),
+            (U8(l), U8(r)) => Bool(l.lt(r.deref())),
+            (U16(l), U16(r)) => Bool(l.lt(r.deref())),
+            (U32(l), U32(r)) => Bool(l.lt(r.deref())),
+            (U64(l), U64(r)) => Bool(l.lt(r.deref())),
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l.lt(&r.cast_into(l_dtype)),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => l.cast_into(r_dtype).lt(r),
+                (l, r) => unreachable!("{} less than {}", l, r),
+            },
         }
     }
 
     /// Element-wise less-or-equal comparison.
     pub fn lte(&self, other: &Array) -> Array {
         use Array::*;
-        match self {
-            Bool(l) => Bool(l.lte(&other.type_cast())),
-            C32(l) => Bool(l.lte(&other.type_cast())),
-            C64(l) => Bool(l.lte(&other.type_cast())),
-            F32(l) => Bool(l.lte(&other.type_cast())),
-            F64(l) => Bool(l.lte(&other.type_cast())),
-            I16(l) => Bool(l.lte(&other.type_cast())),
-            I32(l) => Bool(l.lte(&other.type_cast())),
-            I64(l) => Bool(l.lte(&other.type_cast())),
-            U8(l) => Bool(l.lte(&other.type_cast())),
-            U16(l) => Bool(l.lte(&other.type_cast())),
-            U32(l) => Bool(l.lte(&other.type_cast())),
-            U64(l) => Bool(l.lte(&other.type_cast())),
+        match (self, other) {
+            (Bool(l), Bool(r)) => Bool(l.lte(r.deref())),
+            (C32(l), C32(r)) => Bool(l.lte(r.deref())),
+            (C64(l), C64(r)) => Bool(l.lte(r.deref())),
+            (F32(l), F32(r)) => Bool(l.lte(r.deref())),
+            (F64(l), F64(r)) => Bool(l.lte(r.deref())),
+            (I16(l), I16(r)) => Bool(l.lte(r.deref())),
+            (I32(l), I32(r)) => Bool(l.lte(r.deref())),
+            (I64(l), I64(r)) => Bool(l.lte(r.deref())),
+            (U8(l), U8(r)) => Bool(l.lte(r.deref())),
+            (U16(l), U16(r)) => Bool(l.lte(r.deref())),
+            (U32(l), U32(r)) => Bool(l.lte(r.deref())),
+            (U64(l), U64(r)) => Bool(l.lte(r.deref())),
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l.lte(&r.cast_into(l_dtype)),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => l.cast_into(r_dtype).lte(r),
+                (l, r) => unreachable!("{} less than or equal to {}", l, r),
+            },
         }
     }
 
     /// Element-wise inequality comparison.
     pub fn ne(&self, other: &Array) -> Array {
         use Array::*;
-
-        match self {
-            Bool(l) => Bool(l.ne(&other.type_cast())),
-            C32(l) => Bool(l.ne(&other.type_cast())),
-            C64(l) => Bool(l.ne(&other.type_cast())),
-            F32(l) => Bool(l.ne(&other.type_cast())),
-            F64(l) => Bool(l.ne(&other.type_cast())),
-            I16(l) => Bool(l.ne(&other.type_cast())),
-            I32(l) => Bool(l.ne(&other.type_cast())),
-            I64(l) => Bool(l.ne(&other.type_cast())),
-            U8(l) => Bool(l.ne(&other.type_cast())),
-            U16(l) => Bool(l.ne(&other.type_cast())),
-            U32(l) => Bool(l.ne(&other.type_cast())),
-            U64(l) => Bool(l.ne(&other.type_cast())),
+        match (self, other) {
+            (Bool(l), Bool(r)) => Bool(l.ne(r.deref())),
+            (C32(l), C32(r)) => Bool(l.ne(r.deref())),
+            (C64(l), C64(r)) => Bool(l.ne(r.deref())),
+            (F32(l), F32(r)) => Bool(l.ne(r.deref())),
+            (F64(l), F64(r)) => Bool(l.ne(r.deref())),
+            (I16(l), I16(r)) => Bool(l.ne(r.deref())),
+            (I32(l), I32(r)) => Bool(l.ne(r.deref())),
+            (I64(l), I64(r)) => Bool(l.ne(r.deref())),
+            (U8(l), U8(r)) => Bool(l.ne(r.deref())),
+            (U16(l), U16(r)) => Bool(l.ne(r.deref())),
+            (U32(l), U32(r)) => Bool(l.ne(r.deref())),
+            (U64(l), U64(r)) => Bool(l.ne(r.deref())),
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l.ne(&r.cast_into(l_dtype)),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => l.cast_into(r_dtype).ne(r),
+                (l, r) => unreachable!("{} not equal to {}", l, r),
+            },
         }
     }
 
@@ -498,7 +545,7 @@ impl Array {
     /// Get the values at the specified coordinates.
     pub fn get(&self, index: &ArrayExt<u64>) -> Self {
         let mut indexer = af::Indexer::default();
-        indexer.set_index(index.af(), 0, None);
+        indexer.set_index(index.deref(), 0, None);
         self.get_at(indexer)
     }
 
@@ -522,17 +569,12 @@ impl Array {
 
     /// Return this `Array` raised to the power of `other`.
     pub fn pow(&self, other: &Self) -> Self {
-        if self.dtype() != other.dtype() {
-            let dtype = Ord::max(self.dtype(), other.dtype());
-            return self.cast_into(dtype).pow(&other.cast_into(dtype));
-        }
-
         // af::pow only works with floating point numbers!
-        // raising an f32 af::Array to a power causes a stack overflow!
         use Array::*;
         match (self, other) {
+            (C32(l), C32(r)) => C32(l.pow(r)),
             (C64(l), C64(r)) => C64(l.pow(r)),
-            (C32(l), C32(r)) => C64(l.type_cast()).pow(&C64(r.type_cast())),
+            (F32(l), F32(r)) => F32(l.pow(r)),
             (F64(l), F64(r)) => F64(l.pow(r)),
             (l, r) => {
                 let l = F64(l.type_cast());
@@ -545,7 +587,7 @@ impl Array {
     /// Set the values at the specified coordinates to the corresponding values in `other`.
     pub fn set(&mut self, index: &ArrayExt<u64>, other: &Array) -> Result<()> {
         let mut indexer = af::Indexer::default();
-        indexer.set_index(index.af(), 0, None);
+        indexer.set_index(index.deref(), 0, None);
         self.set_at(indexer, other)
     }
 
@@ -760,33 +802,9 @@ impl PartialEq for Array {
     fn eq(&self, other: &Array) -> bool {
         if self.len() != other.len() {
             return false;
+        } else {
+            Array::eq(self, other).all()
         }
-
-        use Array::*;
-        let eq = match (self, other) {
-            (Bool(l), Bool(r)) => l.eq(r),
-            (C32(l), C32(r)) => l.eq(r),
-            (C64(l), C64(r)) => l.eq(r),
-            (F32(l), F32(r)) => l.eq(r),
-            (F64(l), F64(r)) => l.eq(r),
-            (I16(l), I16(r)) => l.eq(r),
-            (I32(l), I32(r)) => l.eq(r),
-            (I64(l), I64(r)) => l.eq(r),
-            (U8(l), U8(r)) => l.eq(r),
-            (U16(l), U16(r)) => l.eq(r),
-            (U32(l), U32(r)) => l.eq(r),
-            (U64(l), U64(r)) => l.eq(r),
-            (l, r) if l.dtype() > r.dtype() => {
-                let r = r.cast_into(l.dtype());
-                return l == &r;
-            }
-            (l, r) => {
-                let l = l.cast_into(r.dtype());
-                return &l == r;
-            }
-        };
-
-        eq.all()
     }
 }
 
@@ -797,11 +815,9 @@ impl Add for &Array {
         use Array::*;
         match (self, other) {
             (Bool(l), Bool(r)) => Bool(l + r),
-            // Adding an F32 array causes a stack overflow
-            (C32(l), C32(r)) => &C64(l.type_cast()) + &C64(r.type_cast()),
+            (C32(l), C32(r)) => C32(l + r),
             (C64(l), C64(r)) => C64(l + r),
-            // Adding an F32 array causes a stack overflow
-            (F32(l), F32(r)) => &F64(l.type_cast()) + &F64(r.type_cast()),
+            (F32(l), F32(r)) => F32(l + r),
             (F64(l), F64(r)) => F64(l + r),
             (I16(l), I16(r)) => I16(l + r),
             (I32(l), I32(r)) => I32(l + r),
@@ -810,23 +826,61 @@ impl Add for &Array {
             (U16(l), U16(r)) => U16(l + r),
             (U32(l), U32(r)) => U32(l + r),
             (U64(l), U64(r)) => U64(l + r),
-            (l, r) => {
-                let dtype = Ord::max(l.dtype(), r.dtype());
-                let l = l.cast_into(dtype);
-                let r = r.cast_into(dtype);
-                &l + &r
-            }
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l + &r.cast_into(l_dtype),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => &l.cast_into(r_dtype) + r,
+                (l, r) => unreachable!("add {}, {}", l, r),
+            },
         }
     }
 }
 
-impl AddAssign for Array {
-    fn add_assign(&mut self, other: Self) {
-        let sum = &*self + &other;
+impl Add<Number> for &Array {
+    type Output = Array;
+
+    fn add(self, rhs: Number) -> Self::Output {
+        use number_general::Complex;
+        match (self, rhs) {
+            (Array::Bool(l), Number::Bool(r)) => Array::Bool((l.deref() + bool::from(r)).into()),
+
+            (Array::F32(l), Number::Float(Float::F32(r))) => Array::F32((l.deref() + r).into()),
+            (Array::F64(l), Number::Float(Float::F32(r))) => Array::F64((l.deref() + r).into()),
+            (Array::F64(l), Number::Float(Float::F64(r))) => Array::F64((l.deref() + r).into()),
+
+            (Array::C32(l), Number::Complex(Complex::C32(r))) => Array::C32((l.deref() + r).into()),
+            (Array::C64(l), Number::Complex(Complex::C64(r))) => Array::C64((l.deref() + r).into()),
+
+            (Array::I16(l), Number::Int(Int::I16(r))) => Array::I16((l.deref() + r).into()),
+            (Array::I32(l), Number::Int(Int::I32(r))) => Array::I32((l.deref() + r).into()),
+            (Array::I64(l), Number::Int(Int::I64(r))) => Array::I64((l.deref() + r).into()),
+
+            (Array::U8(l), Number::UInt(UInt::U8(r))) => Array::U8((l.deref() + r).into()),
+            (Array::U16(l), Number::UInt(UInt::U16(r))) => Array::U16((l.deref() + r).into()),
+            (Array::U32(l), Number::UInt(UInt::U32(r))) => Array::U32((l.deref() + r).into()),
+            (Array::U64(l), Number::UInt(UInt::U64(r))) => Array::U64((l.deref() + r).into()),
+
+            (l, r) => match (l.dtype(), r.class()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l + r.into_type(l_dtype),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => &l.cast_into(r_dtype) + r,
+                (l, r) => unreachable!("add {}, {}", l, r),
+            },
+        }
+    }
+}
+
+impl AddAssign<&Array> for Array {
+    fn add_assign(&mut self, other: &Array) {
+        let sum = &*self + other;
         *self = sum;
     }
 }
 
+impl AddAssign<Number> for Array {
+    fn add_assign(&mut self, rhs: Number) {
+        *self = &*self + rhs;
+    }
+}
+
 impl Sub for &Array {
     type Output = Array;
 
@@ -834,11 +888,9 @@ impl Sub for &Array {
         use Array::*;
         match (self, other) {
             (Bool(l), Bool(r)) => Bool(l - r),
-            // Subtracting a 32-bit float causes a stack overflow
-            (C32(l), C32(r)) => &C64(l.type_cast()) - &C64(r.type_cast()),
+            (C32(l), C32(r)) => C32(l - r),
             (C64(l), C64(r)) => C64(l - r),
-            // Subtracting a 32-bit float causes a stack overflow
-            (F32(l), F32(r)) => &F64(l.type_cast()) - &F64(r.type_cast()),
+            (F32(l), F32(r)) => F32(l - r),
             (F64(l), F64(r)) => F64(l - r),
             (I16(l), I16(r)) => I16(l - r),
             (I32(l), I32(r)) => I32(l - r),
@@ -847,23 +899,60 @@ impl Sub for &Array {
             (U16(l), U16(r)) => U16(l - r),
             (U32(l), U32(r)) => U32(l - r),
             (U64(l), U64(r)) => U64(l - r),
-            (l, r) => {
-                let dtype = Ord::max(l.dtype(), r.dtype());
-                let l = l.cast_into(dtype);
-                let r = r.cast_into(dtype);
-                &l - &r
-            }
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l - &r.cast_into(l_dtype),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => &l.cast_into(r_dtype) - r,
+                (l, r) => unreachable!("subtract {}, {}", l, r),
+            },
         }
     }
 }
 
-impl SubAssign for Array {
-    fn sub_assign(&mut self, other: Self) {
-        let diff = &*self - &other;
+impl Sub<Number> for &Array {
+    type Output = Array;
+
+    fn sub(self, rhs: Number) -> Self::Output {
+        use number_general::Complex;
+        match (self, rhs) {
+            (Array::Bool(l), Number::Bool(r)) => Array::Bool((l.deref() - bool::from(r)).into()),
+
+            (Array::F32(l), Number::Float(Float::F32(r))) => Array::F32((l.deref() - r).into()),
+            (Array::F64(l), Number::Float(Float::F64(r))) => Array::F64((l.deref() - r).into()),
+
+            (Array::C32(l), Number::Complex(Complex::C32(r))) => Array::C32((l.deref() - r).into()),
+            (Array::C64(l), Number::Complex(Complex::C64(r))) => Array::C64((l.deref() - r).into()),
+
+            (Array::I16(l), Number::Int(Int::I16(r))) => Array::I16((l.deref() - r).into()),
+            (Array::I32(l), Number::Int(Int::I32(r))) => Array::I32((l.deref() - r).into()),
+            (Array::I64(l), Number::Int(Int::I64(r))) => Array::I64((l.deref() - r).into()),
+
+            (Array::U8(l), Number::UInt(UInt::U8(r))) => Array::U8((l.deref() - r).into()),
+            (Array::U16(l), Number::UInt(UInt::U16(r))) => Array::U16((l.deref() - r).into()),
+            (Array::U32(l), Number::UInt(UInt::U32(r))) => Array::U32((l.deref() - r).into()),
+            (Array::U64(l), Number::UInt(UInt::U64(r))) => Array::U64((l.deref() - r).into()),
+
+            (l, r) => match (l.dtype(), r.class()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l - r.into_type(l_dtype),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => &l.cast_into(r_dtype) - r,
+                (l, r) => unreachable!("subtract {}, {}", l, r),
+            },
+        }
+    }
+}
+
+impl SubAssign<&Array> for Array {
+    fn sub_assign(&mut self, other: &Array) {
+        let diff = &*self - other;
         *self = diff;
     }
 }
 
+impl SubAssign<Number> for Array {
+    fn sub_assign(&mut self, rhs: Number) {
+        *self = &*self - rhs;
+    }
+}
+
 impl Mul for &Array {
     type Output = Array;
 
@@ -884,23 +973,60 @@ impl Mul for &Array {
             (U16(l), U16(r)) => U16(l * r),
             (U32(l), U32(r)) => U32(l * r),
             (U64(l), U64(r)) => U64(l * r),
-            (l, r) => {
-                let dtype = Ord::max(l.dtype(), r.dtype());
-                let l = l.cast_into(dtype);
-                let r = r.cast_into(dtype);
-                &l * &r
-            }
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l * &r.cast_into(l_dtype),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => &l.cast_into(r_dtype) * r,
+                (l, r) => unreachable!("multiply {}, {}", l, r),
+            },
         }
     }
 }
 
-impl MulAssign for Array {
-    fn mul_assign(&mut self, other: Array) {
-        let product = &*self * &other;
+impl Mul<Number> for &Array {
+    type Output = Array;
+
+    fn mul(self, rhs: Number) -> Self::Output {
+        use number_general::Complex;
+        match (self, rhs) {
+            (Array::Bool(l), Number::Bool(r)) => Array::Bool((l.deref() * bool::from(r)).into()),
+
+            (Array::F32(l), Number::Float(Float::F32(r))) => Array::F32((l.deref() * r).into()),
+            (Array::F64(l), Number::Float(Float::F64(r))) => Array::F64((l.deref() * r).into()),
+
+            (Array::C32(l), Number::Complex(Complex::C32(r))) => Array::C32((l.deref() * r).into()),
+            (Array::C64(l), Number::Complex(Complex::C64(r))) => Array::C64((l.deref() * r).into()),
+
+            (Array::I16(l), Number::Int(Int::I16(r))) => Array::I16((l.deref() * r).into()),
+            (Array::I32(l), Number::Int(Int::I32(r))) => Array::I32((l.deref() * r).into()),
+            (Array::I64(l), Number::Int(Int::I64(r))) => Array::I64((l.deref() * r).into()),
+
+            (Array::U8(l), Number::UInt(UInt::U8(r))) => Array::U8((l.deref() * r).into()),
+            (Array::U16(l), Number::UInt(UInt::U16(r))) => Array::U16((l.deref() * r).into()),
+            (Array::U32(l), Number::UInt(UInt::U32(r))) => Array::U32((l.deref() * r).into()),
+            (Array::U64(l), Number::UInt(UInt::U64(r))) => Array::U64((l.deref() * r).into()),
+
+            (l, r) => match (l.dtype(), r.class()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l * r.into_type(l_dtype),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => &l.cast_into(r_dtype) * r,
+                (l, r) => unreachable!("subtract {}, {}", l, r),
+            },
+        }
+    }
+}
+
+impl MulAssign<&Array> for Array {
+    fn mul_assign(&mut self, other: &Array) {
+        let product = &*self * other;
         *self = product;
     }
 }
 
+impl MulAssign<Number> for Array {
+    fn mul_assign(&mut self, rhs: Number) {
+        *self = &*self * rhs;
+    }
+}
+
 impl Div for &Array {
     type Output = Array;
 
@@ -908,11 +1034,9 @@ impl Div for &Array {
         use Array::*;
         match (self, other) {
             (Bool(l), Bool(r)) => Bool(l / r),
-            // dividing an F32 array causes a stack overflow
-            (C32(l), C32(r)) => &C64(l.type_cast()) / &C64(r.type_cast()),
+            (C32(l), C32(r)) => C32(l / r),
             (C64(l), C64(r)) => C64(l / r),
-            // dividing an F32 array causes a stack overflow
-            (F32(l), F32(r)) => &F64(l.type_cast()) / &F64(r.type_cast()),
+            (F32(l), F32(r)) => F32(l / r),
             (F64(l), F64(r)) => F64(l / r),
             (I16(l), I16(r)) => I16(l / r),
             (I32(l), I32(r)) => I32(l / r),
@@ -921,22 +1045,60 @@ impl Div for &Array {
             (U16(l), U16(r)) => U16(l / r),
             (U32(l), U32(r)) => U32(l / r),
             (U64(l), U64(r)) => U64(l / r),
-            (l, r) => {
-                let dtype = Ord::max(l.dtype(), r.dtype());
-                let l = l.cast_into(dtype);
-                let r = r.cast_into(dtype);
-                &l / &r
-            }
+            (l, r) => match (l.dtype(), r.dtype()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l / &r.cast_into(l_dtype),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => &l.cast_into(r_dtype) / r,
+                (l, r) => unreachable!("divide {}, {}", l, r),
+            },
         }
     }
 }
 
-impl DivAssign for Array {
-    fn div_assign(&mut self, other: Array) {
-        let div = &*self / &other;
+impl Div<Number> for &Array {
+    type Output = Array;
+
+    fn div(self, rhs: Number) -> Self::Output {
+        use number_general::Complex;
+        match (self, rhs) {
+            (Array::Bool(l), Number::Bool(r)) => Array::Bool((l.deref() / bool::from(r)).into()),
+
+            (Array::F32(l), Number::Float(Float::F32(r))) => Array::F32((l.deref() / r).into()),
+            (Array::F64(l), Number::Float(Float::F64(r))) => Array::F64((l.deref() / r).into()),
+
+            (Array::C32(l), Number::Complex(Complex::C32(r))) => Array::C32((l.deref() / r).into()),
+            (Array::C64(l), Number::Complex(Complex::C64(r))) => Array::C64((l.deref() / r).into()),
+
+            (Array::I16(l), Number::Int(Int::I16(r))) => Array::I16((l.deref() / r).into()),
+            (Array::I32(l), Number::Int(Int::I32(r))) => Array::I32((l.deref() / r).into()),
+            (Array::I64(l), Number::Int(Int::I64(r))) => Array::I64((l.deref() / r).into()),
+
+            (Array::U8(l), Number::UInt(UInt::U8(r))) => Array::U8((l.deref() / r).into()),
+            (Array::U16(l), Number::UInt(UInt::U16(r))) => Array::U16((l.deref() / r).into()),
+            (Array::U32(l), Number::UInt(UInt::U32(r))) => Array::U32((l.deref() / r).into()),
+            (Array::U64(l), Number::UInt(UInt::U64(r))) => Array::U64((l.deref() / r).into()),
+
+            (l, r) => match (l.dtype(), r.class()) {
+                (l_dtype, r_dtype) if l_dtype > r_dtype => l / r.into_type(l_dtype),
+                (l_dtype, r_dtype) if l_dtype < r_dtype => &l.cast_into(r_dtype) / r,
+                (l, r) => unreachable!("subtract {}, {}", l, r),
+            },
+        }
+    }
+}
+
+impl DivAssign<&Array> for Array {
+    fn div_assign(&mut self, other: &Array) {
+        let div = &*self / other;
         *self = div;
     }
 }
+
+impl DivAssign<Number> for Array {
+    fn div_assign(&mut self, rhs: Number) {
+        *self = &*self / rhs;
+    }
+}
+
 impl<T: af::HasAfEnum> CastFrom<Array> for ArrayExt<T> {
     fn cast_from(array: Array) -> ArrayExt<T> {
         use Array::*;
@@ -1324,6 +1486,7 @@ mod tests {
         actual
             .set(&(&[1, 2][..]).into(), &Array::from(&[4, 5][..]))
             .unwrap();
+
         let expected = Array::from(&[1, 4, 5][..]);
         assert_eq!(actual, expected)
     }
@@ -1336,6 +1499,8 @@ mod tests {
 
         let b: Array = [3, 2, 1][..].into();
         assert_eq!(&a + &b, [4, 4, 4][..].into());
+
+        assert_eq!(&b + Number::from(1), [4, 3, 2][..].into());
     }
 
     #[test]
@@ -1346,6 +1511,8 @@ mod tests {
 
         let b: Array = [-1., -4., 4.][..].into();
         assert_eq!(&a + &b, [0., -2., 7.][..].into());
+
+        assert_eq!(&b + Number::from(3), [2, -1, 7][..].into());
     }
 
     #[test]
