@@ -13,39 +13,33 @@ use serde::ser::{Serialize, Serializer};
 use super::{dim4, Complex};
 
 /// Defines common access methods for instance of [`ArrayExt`].
-pub trait ArrayInstance {
+pub trait ArrayInstance: Deref<Target = af::Array<Self::DType>> + DerefMut {
     type DType: af::HasAfEnum;
 
-    /// Borrow this instance as an [`af::Array`].
-    fn af(&'_ self) -> &'_ af::Array<Self::DType>;
-
-    /// Borrow this instance as a a mutable [`af::Array`].
-    fn af_mut(&'_ mut self) -> &'_ mut af::Array<Self::DType>;
-
     /// Cast this instance into an [`af::Array`] with type `T`.
     fn af_cast<T: af::HasAfEnum>(&self) -> af::Array<T> {
-        self.af().cast()
+        self.cast()
     }
 
     /// How many elements are in this `ArrayInstance`.
     fn len(&self) -> usize {
-        self.af().elements()
+        self.elements()
     }
 
     /// Get the values specified by the given [`af::Indexer`].
     fn get(&self, index: af::Indexer) -> af::Array<Self::DType> {
-        af::index_gen(self.af(), index)
+        af::index_gen(self, index)
     }
 
     /// Set the values specified by the given [`af::Indexer`] to the corresponding values in `T`.
     fn set<T: ArrayInstance<DType = Self::DType>>(&mut self, index: &af::Indexer, other: &T) {
-        af::assign_gen(self.af_mut(), index, other.af());
+        af::assign_gen(self, index, other);
     }
 
     /// Set the value at the specified index to `value`.
     fn set_at(&mut self, offset: usize, value: Self::DType) {
         let seq = af::Seq::new(offset as f32, offset as f32, 1.0f32);
-        af::assign_seq(self.af_mut(), &[seq], &af::Array::new(&[value], dim4(1)));
+        af::assign_seq(self, &[seq], &af::Array::new(&[value], dim4(1)));
     }
 
     /// Copy the data in this [`af::Array`] into a new `Vec`.
@@ -54,7 +48,7 @@ pub trait ArrayInstance {
         Self::DType: Clone + Default,
     {
         let mut v = vec![Self::DType::default(); self.len()];
-        self.af().host(&mut v);
+        self.host(&mut v);
         v
     }
 }
@@ -75,7 +69,7 @@ impl<T: af::HasAfEnum + Default> ArrayExt<T> {
 
     /// Concatenate two instances of `ArrayExt<T>`.
     pub fn concatenate(left: &Self, right: &Self) -> Self {
-        af::join(0, left.af(), right.af()).into()
+        af::join(0, left, right).into()
     }
 
     /// Return `true` if the elements of this `ArrayExt` are in sorted order.
@@ -83,7 +77,8 @@ impl<T: af::HasAfEnum + Default> ArrayExt<T> {
     where
         T: af::RealNumber + Clone,
     {
-        af::all_true_all(&af::eq(self.af(), self.sorted(true).af(), false)).0
+        let sorted = self.sorted(true);
+        af::all_true_all(&af::eq(self.deref(), sorted.deref(), false)).0
     }
 
     /// Cast the values of this `ArrayExt` into a destination type `D`.
@@ -98,9 +93,9 @@ impl<T: af::HasAfEnum + Default> ArrayExt<T> {
 
     /// Get the value at the given index.
     pub fn get_value(&self, index: usize) -> T {
-        debug_assert!(index < self.af().elements());
+        debug_assert!(index < self.elements());
 
-        let af_value = af::index(self.af(), &[af::Seq::new(index as f64, index as f64, 1f64)]);
+        let af_value = af::index(self, &[af::Seq::new(index as f64, index as f64, 1f64)]);
         if af_value.elements() == 1 {
             let mut value = vec![T::default()];
             af_value.host(&mut value);
@@ -119,11 +114,7 @@ impl<T: af::HasAfEnum + Default> ArrayExt<T> {
     ///
     /// Panics: if `end` is out of bounds
     pub fn slice(&self, start: usize, end: usize) -> Self {
-        af::index(
-            self.af(),
-            &[af::Seq::new(start as f32, (end - 1) as f32, 1.)],
-        )
-        .into()
+        af::index(self, &[af::Seq::new(start as f32, (end - 1) as f32, 1.)]).into()
     }
 
     /// Sort this `ArrayExt`.
@@ -148,8 +139,8 @@ impl<T: af::HasAfEnum + Default> ArrayExt<T> {
         let left = af::Seq::new(0.0, at as f32, 1.0);
         let right = af::Seq::new(at as f32, self.len() as f32, 1.0);
         (
-            ArrayExt(af::index(self.af(), &[left, af::Seq::default()])),
-            ArrayExt(af::index(self.af(), &[right, af::Seq::default()])),
+            ArrayExt(af::index(self, &[left, af::Seq::default()])),
+            ArrayExt(af::index(self, &[right, af::Seq::default()])),
         )
     }
 
@@ -160,7 +151,7 @@ impl<T: af::HasAfEnum + Default> ArrayExt<T> {
     where
         T: af::RealNumber,
     {
-        Self(af::set_unique(self.af(), sorted))
+        Self(af::set_unique(self, sorted))
     }
 
     fn into_stream(self) -> impl Stream<Item = Vec<T>>
@@ -178,27 +169,45 @@ impl<T: af::HasAfEnum + Default> ArrayExt<T> {
     }
 }
 
+impl<T: af::HasAfEnum> ArrayInstance for ArrayExt<T> {
+    type DType = T;
+}
+
+impl<T: af::HasAfEnum> Deref for ArrayExt<T> {
+    type Target = af::Array<T>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.0
+    }
+}
+
+impl<T: af::HasAfEnum> DerefMut for ArrayExt<T> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.0
+    }
+}
+
 impl ArrayExt<bool> {
     /// Logical not.
     pub fn not(&self) -> Self {
-        ArrayExt(!self.af())
+        ArrayExt(!self.deref())
     }
 
     /// Logical and.
     pub fn and(&self, other: &Self) -> Self {
-        ArrayExt(af::and(self.af(), other.af(), batch(self, other)))
+        ArrayExt(af::and(self, other, batch(self, other)))
     }
 
     /// Logical or.
     pub fn or(&self, other: &Self) -> Self {
-        ArrayExt(af::or(self.af(), other.af(), batch(self, other)))
+        ArrayExt(af::or(self, other, batch(self, other)))
     }
 
     /// Logical xor.
     pub fn xor(&self, other: &Self) -> Self {
         let batch = batch(self, other);
-        let one = af::or(self.af(), other.af(), batch);
-        let not_both = !(&af::and(self.af(), other.af(), batch));
+        let one = af::or(self, other, batch);
+        let not_both = !(&af::and(self, other, batch));
         let one_and_not_both = af::and(&one, &not_both, batch);
         ArrayExt(one_and_not_both)
     }
@@ -242,20 +251,6 @@ impl ArrayExt<u64> {
     }
 }
 
-impl<T: af::HasAfEnum> ArrayInstance for ArrayExt<T> {
-    type DType = T;
-
-    #[inline]
-    fn af(&'_ self) -> &'_ af::Array<T> {
-        &self.0
-    }
-
-    #[inline]
-    fn af_mut(&mut self) -> &mut af::Array<Self::DType> {
-        &mut self.0
-    }
-}
-
 impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> Add for ArrayExt<T>
     where
         <T as af::ImplicitPromote<T>>::Output: af::HasAfEnum + Default,
@@ -282,18 +277,6 @@ impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> A
     }
 }
 
-impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> AddAssign for ArrayExt<T>
-    where
-        <T as af::ImplicitPromote<T>>::Output: af::HasAfEnum + Default,
-        <T as af::Convertable>::OutType: af::ImplicitPromote<<T as af::Convertable>::OutType>,
-        <<T as af::Convertable>::OutType as af::ImplicitPromote<<T as af::Convertable>::OutType>>::Output: af::HasAfEnum, {
-
-    fn add_assign(&mut self, other: Self) {
-        let sum = &*self + &other;
-        *self = sum.type_cast();
-    }
-}
-
 impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> Mul for ArrayExt<T>
     where
         <T as af::ImplicitPromote<T>>::Output: af::HasAfEnum + Default,
@@ -320,18 +303,6 @@ impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> M
     }
 }
 
-impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> MulAssign for ArrayExt<T>
-    where
-        <T as af::ImplicitPromote<T>>::Output: af::HasAfEnum + Default,
-        <T as af::Convertable>::OutType: af::ImplicitPromote<<T as af::Convertable>::OutType>,
-        <<T as af::Convertable>::OutType as af::ImplicitPromote<<T as af::Convertable>::OutType>>::Output: af::HasAfEnum, {
-
-    fn mul_assign(&mut self, other: Self) {
-        let product = &*self * &other;
-        *self = product.type_cast();
-    }
-}
-
 impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> Div for ArrayExt<T>
     where
         <T as af::ImplicitPromote<T>>::Output: af::HasAfEnum + Default,
@@ -358,6 +329,19 @@ impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> D
     }
 }
 
+impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> DivAssign for ArrayExt<T>
+where
+    <T as af::ImplicitPromote<T>>::Output: af::HasAfEnum + Default,
+    <T as af::Convertable>::OutType: af::ImplicitPromote<<T as af::Convertable>::OutType>,
+    <<T as af::Convertable>::OutType as af::ImplicitPromote<<T as af::Convertable>::OutType>>::Output: af::HasAfEnum
+{
+
+    fn div_assign(&mut self, other: Self) {
+        let quotient = &*self / &other;
+        *self = quotient.type_cast();
+    }
+}
+
 impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> Sub for ArrayExt<T>
     where
         <T as af::ImplicitPromote<T>>::Output: af::HasAfEnum + Default,
@@ -425,14 +409,14 @@ impl<T: af::HasAfEnum + af::ImplicitPromote<T> + af::Convertable<OutType = T>> R
 impl From<(ArrayExt<f32>, ArrayExt<f32>)> for ArrayExt<Complex<f32>> {
     fn from(elements: (ArrayExt<f32>, ArrayExt<f32>)) -> Self {
         let (re, im) = elements;
-        Self(af::cplx2(re.af(), im.af(), false).cast())
+        Self(af::cplx2(&*re, &*im, false).cast())
     }
 }
 
 impl From<(ArrayExt<f64>, ArrayExt<f64>)> for ArrayExt<Complex<f64>> {
     fn from(elements: (ArrayExt<f64>, ArrayExt<f64>)) -> Self {
         let (re, im) = elements;
-        Self(af::cplx2(re.af(), im.af(), false).cast())
+        Self(af::cplx2(&*re, &*im, false).cast())
     }
 }
 
@@ -475,7 +459,7 @@ impl ArrayInstanceAbs for ArrayExt<Complex<f32>> {
     type AbsValue = f32;
 
     fn abs(&self) -> ArrayExt<f32> {
-        ArrayExt(af::abs(self.af()))
+        ArrayExt(af::abs(&*self))
     }
 }
 
@@ -483,7 +467,7 @@ impl ArrayInstanceAbs for ArrayExt<Complex<f64>> {
     type AbsValue = f64;
 
     fn abs(&self) -> ArrayExt<f64> {
-        ArrayExt(af::abs(self.af()))
+        ArrayExt(af::abs(&*self))
     }
 }
 
@@ -491,7 +475,7 @@ impl ArrayInstanceAbs for ArrayExt<f32> {
     type AbsValue = f32;
 
     fn abs(&self) -> ArrayExt<f32> {
-        ArrayExt(af::abs(self.af()))
+        ArrayExt(af::abs(&*self))
     }
 }
 
@@ -499,7 +483,7 @@ impl ArrayInstanceAbs for ArrayExt<f64> {
     type AbsValue = f64;
 
     fn abs(&self) -> ArrayExt<f64> {
-        ArrayExt(af::abs(self.af()))
+        ArrayExt(af::abs(&*self))
     }
 }
 
@@ -507,7 +491,7 @@ impl ArrayInstanceAbs for ArrayExt<i16> {
     type AbsValue = i16;
 
     fn abs(&self) -> ArrayExt<i16> {
-        ArrayExt(af::abs(self.af()).cast())
+        ArrayExt(af::abs(&*self).cast())
     }
 }
 
@@ -515,7 +499,7 @@ impl ArrayInstanceAbs for ArrayExt<i32> {
     type AbsValue = i32;
 
     fn abs(&self) -> ArrayExt<i32> {
-        ArrayExt(af::abs(self.af()).cast())
+        ArrayExt(af::abs(&*self).cast())
     }
 }
 
@@ -523,7 +507,7 @@ impl ArrayInstanceAbs for ArrayExt<i64> {
     type AbsValue = i64;
 
     fn abs(&self) -> ArrayExt<i64> {
-        ArrayExt(af::abs(self.af()).cast())
+        ArrayExt(af::abs(&*self).cast())
     }
 }
 
@@ -531,12 +515,12 @@ impl ArrayInstanceAbs for ArrayExt<i64> {
 pub trait ArrayInstanceAnyAll: ArrayInstance {
     /// Returns `true` if all elements are nonzero.
     fn all(&self) -> bool {
-        af::all_true_all(self.af()).0
+        af::all_true_all(&*self).0
     }
 
     /// Returns `true` if any element is nonzero.
     fn any(&self) -> bool {
-        af::any_true_all(self.af()).0
+        af::any_true_all(&*self).0
     }
 }
 
@@ -553,91 +537,105 @@ impl ArrayInstanceAnyAll for ArrayExt<u64> {}
 
 impl ArrayInstanceAnyAll for ArrayExt<Complex<f32>> {
     fn all(&self) -> bool {
-        af::all_true_all(&af::abs(self.af())).0
+        af::all_true_all(&af::abs(&*self)).0
     }
 
     fn any(&self) -> bool {
-        let any = af::any_true_all(self.af());
+        let any = af::any_true_all(&*self);
         any.0 || any.1
     }
 }
 
 impl ArrayInstanceAnyAll for ArrayExt<Complex<f64>> {
     fn all(&self) -> bool {
-        af::all_true_all(&af::abs(self.af())).0
+        af::all_true_all(&af::abs(&*self)).0
     }
 
     fn any(&self) -> bool {
-        let any = af::any_true_all(self.af());
+        let any = af::any_true_all(&*self);
         any.0 || any.1
     }
 }
 
-/// Defines element-wise comparison operations.
-pub trait ArrayInstanceCompare {
+pub trait ArrayInstanceUnreal {
     /// Element-wise check for infinite values.
     fn is_infinite(&self) -> ArrayExt<bool>;
 
     /// Element-wise check for non-numeric (NaN) values.
     fn is_nan(&self) -> ArrayExt<bool>;
-
-    /// Element-wise equality.
-    fn eq(&self, other: &Self) -> ArrayExt<bool>;
-
-    /// Element-wise greater-than comparison.
-    fn gt(&self, other: &Self) -> ArrayExt<bool>;
-
-    /// Element-wise greater-or-equal comparison
-    fn gte(&self, other: &Self) -> ArrayExt<bool>;
-
-    /// Element-wise less-than comparison.
-    fn lt(&self, other: &Self) -> ArrayExt<bool>;
-
-    /// Element-wise less-or-equal comparison.
-    fn lte(&self, other: &Self) -> ArrayExt<bool>;
-
-    /// Element-wise inequality.
-    fn ne(&self, other: &Self) -> ArrayExt<bool>;
 }
 
-impl<T: af::HasAfEnum + af::ImplicitPromote<T>> ArrayInstanceCompare for ArrayExt<T> {
+impl<T: af::HasAfEnum + af::ImplicitPromote<T>> ArrayInstanceUnreal for ArrayExt<T> {
     fn is_infinite(&self) -> ArrayExt<bool> {
-        af::isinf(self.af()).into()
+        af::isinf(&*self).into()
     }
 
     fn is_nan(&self) -> ArrayExt<bool> {
-        af::isnan(self.af()).into()
+        af::isnan(&*self).into()
+    }
+}
+
+/// Defines element-wise comparison operations.
+pub trait ArrayInstanceCompare<U> {
+    /// Element-wise equality.
+    fn eq(&self, other: &U) -> ArrayExt<bool>;
+
+    /// Element-wise greater-than comparison.
+    fn gt(&self, other: &U) -> ArrayExt<bool>;
+
+    /// Element-wise greater-or-equal comparison
+    fn gte(&self, other: &U) -> ArrayExt<bool>;
+
+    /// Element-wise less-than comparison.
+    fn lt(&self, other: &U) -> ArrayExt<bool>;
+
+    /// Element-wise less-or-equal comparison.
+    fn lte(&self, other: &U) -> ArrayExt<bool>;
+
+    /// Element-wise inequality.
+    fn ne(&self, other: &U) -> ArrayExt<bool>;
+}
+
+impl<T, U> ArrayInstanceCompare<U> for ArrayExt<T>
+where
+    T: Clone + af::HasAfEnum + af::Convertable<OutType = T>,
+    U: af::Convertable<OutType = T>,
+    <T as af::Convertable>::OutType: af::ImplicitPromote<<U as af::Convertable>::OutType>,
+    <U as af::Convertable>::OutType: af::ImplicitPromote<<T as af::Convertable>::OutType>,
+{
+    fn eq(&self, other: &U) -> ArrayExt<bool> {
+        af::eq(self.deref(), other, true).into()
     }
 
-    fn eq(&self, other: &Self) -> ArrayExt<bool> {
-        af::eq(self.af(), other.af(), batch(self, other)).into()
+    fn gt(&self, other: &U) -> ArrayExt<bool> {
+        af::gt(self.deref(), other, true).into()
     }
 
-    fn gt(&self, other: &Self) -> ArrayExt<bool> {
-        af::gt(self.af(), other.af(), batch(self, other)).into()
+    fn gte(&self, other: &U) -> ArrayExt<bool> {
+        af::ge(self.deref(), other, true).into()
     }
 
-    fn gte(&self, other: &Self) -> ArrayExt<bool> {
-        af::ge(self.af(), other.af(), batch(self, other)).into()
+    fn lt(&self, other: &U) -> ArrayExt<bool> {
+        af::lt(self.deref(), other, true).into()
     }
 
-    fn lt(&self, other: &Self) -> ArrayExt<bool> {
-        af::lt(self.af(), other.af(), batch(self, other)).into()
+    fn lte(&self, other: &U) -> ArrayExt<bool> {
+        af::le(self.deref(), other, true).into()
     }
 
-    fn lte(&self, other: &Self) -> ArrayExt<bool> {
-        af::le(self.af(), other.af(), batch(self, other)).into()
-    }
-
-    fn ne(&self, other: &Self) -> ArrayExt<bool> {
+    fn ne(&self, other: &U) -> ArrayExt<bool> {
         self.eq(other).not()
     }
 }
 
 /// Defines an exponentiation method `pow`.
 pub trait ArrayInstancePow: ArrayInstance {
+    type Exp: af::HasAfEnum;
     type Pow: af::HasAfEnum;
 
+    /// Raise `e` to the power of `self`.
+    fn exp(&self) -> ArrayExt<Self::Exp>;
+
     /// Calculate the element-wise exponentiation.
     fn pow(&self, other: &Self) -> ArrayExt<Self::Pow>;
 }
@@ -645,13 +643,19 @@ pub trait ArrayInstancePow: ArrayInstance {
 impl<T> ArrayInstancePow for ArrayExt<T>
 where
     T: af::HasAfEnum + af::ConstGenerator<OutType = T> + Clone + Default,
+    T::UnaryOutType: af::HasAfEnum,
     <T as af::Convertable>::OutType: af::ImplicitPromote<<T as af::Convertable>::OutType>,
     <<T as af::Convertable>::OutType as af::ImplicitPromote<<T as af::Convertable>::OutType>>::Output: af::HasAfEnum,
 {
+    type Exp = T::UnaryOutType;
     type Pow = <<T as af::Convertable>::OutType as af::ImplicitPromote<<T as af::Convertable>::OutType>>::Output;
 
+    fn exp(&self) -> ArrayExt<Self::Exp> {
+        af::exp(self).into()
+    }
+
     fn pow(&self, other: &Self) -> ArrayExt<Self::Pow> {
-        af::pow(self.af(), other.af(), true).into()
+        af::pow(self.deref(), other.deref(), true).into()
     }
 }
 
@@ -672,11 +676,11 @@ impl ArrayInstanceReduce for ArrayExt<bool> {
     type Sum = u64;
 
     fn product(&self) -> u64 {
-        af::product_all(self.af()).0 as u64
+        af::product_all(self).0 as u64
     }
 
     fn sum(&self) -> u64 {
-        af::sum_all(self.af()).0 as u64
+        af::sum_all(self).0 as u64
     }
 }
 
@@ -685,12 +689,12 @@ impl ArrayInstanceReduce for ArrayExt<Complex<f32>> {
     type Sum = Complex<f32>;
 
     fn product(&self) -> Self::Product {
-        let product = af::product_all(self.af());
+        let product = af::product_all(self);
         Complex::new(product.0, product.1)
     }
 
     fn sum(&self) -> Self::Sum {
-        let sum = af::sum_all(self.af());
+        let sum = af::sum_all(self);
         Complex::new(sum.0, sum.1)
     }
 }
@@ -700,12 +704,12 @@ impl ArrayInstanceReduce for ArrayExt<Complex<f64>> {
     type Sum = Complex<f64>;
 
     fn product(&self) -> Self::Product {
-        let product = af::product_all(self.af());
+        let product = af::product_all(self);
         Complex::new(product.0, product.1)
     }
 
     fn sum(&self) -> Self::Sum {
-        let sum = af::sum_all(self.af());
+        let sum = af::sum_all(self);
         Complex::new(sum.0, sum.1)
     }
 }
@@ -715,11 +719,11 @@ impl ArrayInstanceReduce for ArrayExt<f32> {
     type Sum = f32;
 
     fn product(&self) -> f32 {
-        af::product_all(self.af()).0
+        af::product_all(self).0
     }
 
     fn sum(&self) -> f32 {
-        af::sum_all(self.af()).0
+        af::sum_all(self).0
     }
 }
 
@@ -728,11 +732,11 @@ impl ArrayInstanceReduce for ArrayExt<f64> {
     type Sum = f64;
 
     fn product(&self) -> f64 {
-        af::product_all(self.af()).0
+        af::product_all(self).0
     }
 
     fn sum(&self) -> f64 {
-        af::sum_all(self.af()).0
+        af::sum_all(self).0
     }
 }
 
@@ -741,11 +745,11 @@ impl ArrayInstanceReduce for ArrayExt<i16> {
     type Sum = i64;
 
     fn product(&self) -> i64 {
-        af::product_all(self.af()).0 as i64
+        af::product_all(self).0 as i64
     }
 
     fn sum(&self) -> i64 {
-        af::sum_all(self.af()).0 as i64
+        af::sum_all(self).0 as i64
     }
 }
 
@@ -754,11 +758,11 @@ impl ArrayInstanceReduce for ArrayExt<i32> {
     type Sum = i64;
 
     fn product(&self) -> i64 {
-        af::product_all(self.af()).0 as i64
+        af::product_all(self).0 as i64
     }
 
     fn sum(&self) -> i64 {
-        af::sum_all(self.af()).0 as i64
+        af::sum_all(self).0 as i64
     }
 }
 
@@ -767,11 +771,11 @@ impl ArrayInstanceReduce for ArrayExt<i64> {
     type Sum = i64;
 
     fn product(&self) -> i64 {
-        af::product_all(self.af()).0 as i64
+        af::product_all(self).0 as i64
     }
 
     fn sum(&self) -> i64 {
-        af::sum_all(self.af()).0 as i64
+        af::sum_all(self).0 as i64
     }
 }
 
@@ -780,11 +784,11 @@ impl ArrayInstanceReduce for ArrayExt<u8> {
     type Sum = u64;
 
     fn product(&self) -> u64 {
-        af::product_all(self.af()).0 as u64
+        af::product_all(self).0 as u64
     }
 
     fn sum(&self) -> u64 {
-        af::sum_all(self.af()).0 as u64
+        af::sum_all(self).0 as u64
     }
 }
 
@@ -793,11 +797,11 @@ impl ArrayInstanceReduce for ArrayExt<u16> {
     type Sum = u64;
 
     fn product(&self) -> u64 {
-        af::product_all(self.af()).0 as u64
+        af::product_all(self).0 as u64
     }
 
     fn sum(&self) -> u64 {
-        af::sum_all(self.af()).0 as u64
+        af::sum_all(self).0 as u64
     }
 }
 
@@ -806,11 +810,11 @@ impl ArrayInstanceReduce for ArrayExt<u32> {
     type Sum = u64;
 
     fn product(&self) -> u64 {
-        af::product_all(self.af()).0 as u64
+        af::product_all(self).0 as u64
     }
 
     fn sum(&self) -> u64 {
-        af::sum_all(self.af()).0 as u64
+        af::sum_all(self).0 as u64
     }
 }
 
@@ -819,11 +823,11 @@ impl ArrayInstanceReduce for ArrayExt<u64> {
     type Sum = u64;
 
     fn product(&self) -> u64 {
-        af::product_all(self.af()).0 as u64
+        af::product_all(self).0 as u64
     }
 
     fn sum(&self) -> u64 {
-        af::sum_all(self.af()).0 as u64
+        af::sum_all(self).0 as u64
     }
 }
 
@@ -1078,7 +1082,7 @@ impl<T: af::HasAfEnum + fmt::Display + Default + Clone> fmt::Debug for ArrayExt<
             f,
             "ArrayExt<{}>({}): {}",
             std::any::type_name::<T>(),
-            self.af().dims(),
+            self.dims(),
             self
         )
     }
